import contextlib
import time
from functools import wraps

import numpy as np
from csb.numeric import log

from geosss.mcmc import (MetropolisHastings, RejectionSphericalSliceSampler,
                         ShrinkageSphericalSliceSampler, SphericalHMC)

# some nice colors
colors = [(.85, .3, .1), (.15, .35, .6),
          (.95, .7, 0.1), (.0, .0, .0), (.8, .8, .8)]


def format_time(t):

    units = [(1., 's'), (1e-3, 'ms'), (1e-6, 'us'), (1e-9, 'ns')]
    for scale, unit in units:
        if t > scale or t == 0:
            break

    return '{0:.1f} {1}'.format(t/scale, unit)


@contextlib.contextmanager
def take_time(desc, mute=False):
    t0 = time.process_time()
    yield
    dt = time.process_time() - t0
    if not mute:
        print('{0} took {1}'.format(desc, format_time(dt)))


def relative_entropy(p, q):
    """Kullback-Leibler divergence (aka as relative entropy). """
    return p @ (log(p) - log(q))


def acf(x, n_max=None):
    """Autocorrelation.

    Parameters
    ----------
    x : 1d array
    n_max : maximum lag (if None then len(x) // 2)

    Returns
    -------
    Array storing the estimated autocorrelation. 
    """
    n_max = n_max or len(x) // 2
    x = x - np.mean(x)
    ac = [np.mean(x[i:] * x[:len(x) - i]) for i in range(n_max)]
    return np.array(ac) / ac[0]


def acf_fft(x):
    """Compute autocorrelation function using the convolution theorem."""
    X = np.fft.rfft((x - x.mean()) / x.std())
    return np.real(np.fft.irfft(X.conj() * X))[:len(x)//2] / len(x)


def IAT(x, n=None):
    """Computes the integrated autocorrelation time for given values by the heuristics
    described in Gelman et al "Bayesian Data Analysis", Chapter 11.5
    """
    ac = acf_fft(x)
    if n:
        ac = ac[:n]
    sums = ac[2:-1].reshape(-1, 2).sum(1) if (len(ac) % 2 != 0) \
        else ac[2:].reshape(-1, 2).sum(1)
    T = np.nonzero(sums < 0)[0][0]
    L = (1 + 2 * T) if np.sum(sums < 0) else len(ac)-1
    return 1.0 + np.max([2 * np.sum(ac[1:L+1]), 0.0])


def n_eff(x, n=None):
    """Computes effective sample size n_eff for given values
    """
    return len(x) / IAT(x, n)


def count_calls(func):
    """
    decorator that counts how many times a method/function was called. If needed 
    could be called directly only on a function with @count_calls 
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        """ wrapper inside the `count_calls`"""
        # updates the counter everytime func is called
        wrapper.num_calls += 1
        return func(*args, **kwargs)

    # counter starts
    wrapper.num_calls = 0

    def reset_counters():
        """Reset the num_calls counter"""
        wrapper.num_calls = 0

    wrapper.reset_counters = reset_counters

    return wrapper


def counter(method_names):
    """
    A decorator that counts how many times a method was called. It calls the `count_calls` function to 
    wrap around a class. So it should be used only on a class. Example: @counter("some_method") or 
    @counter(["some_method", "another_method"])

    Args:
        method_names (str or list): the name(s) of the method(s) to decorate.

    Returns:
        A class decorator.
    """
    # if argument is a single method as string
    if isinstance(method_names, str):
        method_names = [method_names]

    # `class_decorator` calls `count_calls` decorator
    def class_decorator(cls):
        for method_name in method_names:
            if hasattr(cls, method_name):
                setattr(cls, method_name, count_calls(
                    getattr(cls, method_name)))
        return cls

    return class_decorator


class SamplerLauncher:
    """Just an interface for launching all the samplers
    """

    def __init__(self, pdf, initial, n_samples, burnin=0.2, seed=None):
        self.pdf = pdf
        self.initial = initial
        self.n_samples = n_samples
        self.burnin = burnin
        self.seed = seed

    def run_sss_reject(self):
        sampler = RejectionSphericalSliceSampler(
            self.pdf, self.initial, self.seed)
        self.rsss = sampler

        return sampler.sample(self.n_samples, burnin=self.burnin)

    def run_sss_shrink(self):
        sampler = ShrinkageSphericalSliceSampler(
            self.pdf, self.initial, self.seed)
        self.ssss = sampler

        return sampler.sample(self.n_samples, burnin=self.burnin)

    def run_rwmh(self):
        sampler = MetropolisHastings(
            self.pdf, self.initial, self.seed, stepsize=1e-1)
        self.rwmh = sampler

        return sampler.sample(self.n_samples, burnin=self.burnin)

    def run_hmc(self):
        sampler = SphericalHMC(self.pdf, self.initial,
                               self.seed, stepsize=1e-1)
        self.hmc = sampler

        return sampler.sample(self.n_samples, burnin=self.burnin)

    def run(self, method):
        if method == 'sss-reject':
            return self.run_sss_reject()
        elif method == 'sss-shrink':
            return self.run_sss_shrink()
        elif method == 'rwmh':
            return self.run_rwmh()
        elif method == 'hmc':
            return self.run_hmc()
        else:
            raise ValueError(f'method {method} not known')



"""
General purpose Markov chain Monte Carlo algorithms:
- Random walk Metropolis-Hastings algorithm on the sphere
- Spherical Hamiltonian Monte Carlo
- Slice sampling on the sphere using a rejection and a shrinkage sampler
"""
import numpy as np

from geosss import sphere


def determine_burnin(n_samples, burnin):
    if isinstance(burnin, float):
        assert 0 <= burnin <= 1.
        return int(burnin * n_samples)
    else:
        assert burnin >= 0
        return int(burnin)


class Sampler:
    """Sampler

    Base class for Markov chain Monte Carlo samplers.
    """

    def __init__(self, distribution, initial_state, seed=None, **kwargs):
        """
        Parameters
        ----------
        distribution: instance of distributions.Distribution
            Target distribution implementing `log_prob`, stationary distribution
            of the Markov chain.

        initial_state: numpy ndarray
            Initial state of the Markov chain.

        seed: int, optional
            Seed for the random number generator.
        """
        super().__init__(**kwargs)
        self.target = distribution
        self.state = initial_state
        self.rng = np.random.default_rng(seed)

    def __next__(self):
        """
        Markov chain Monte Carlo sampler is implemented as an `iterator`. Subclasses
        need to implement the `__next__` routine which is supposed to simulate a single
        step of the Markov chain.
        """
        raise NotImplementedError

    def sample(self, n_samples, burnin=0):
        """
        Generates a Markov chain consisting of the desired size. An optional burnin can
        be specified, either as an integer (in which case it is taken as is) or a float
        smaller than one (in which case it is interpreted as percentage of the desired
        number of samples).
        """
        assert n_samples > 0

        burnin = determine_burnin(n_samples, burnin)

        # simulate Markov chain
        samples = [self.state]
        while len(samples) < (n_samples + burnin):
            samples.append(np.copy(next(self)))

        # chop away burn-in and return remaining samples
        return np.array(samples[burnin:])


class AdaptiveStepsize:
    """AdaptiveStepsize

    Mixin for adapting the stepsize in versions of Metropolis-Hastings such as random
    walk MH or spherical Hamiltonian Monte Carlo.

    Attributes
    ----------
    stepsize: float > 0
        Step-size parameter, e.g. variance of an isotropic Gaussian proposal.

    adapt_stepsize: boolean
        Flag indicating if the stepsize should be adapted or not. Typically, this flag
        is switched on in the initial phase of the simulation (burnin) and switched off
        after burn-in in order to ensure detailed balance.
    """

    def __init__(self, stepsize=1e-3, **kwargs):
        super().__init__(**kwargs)
        self.stepsize = float(stepsize)
        assert self.stepsize > 0.
        self.reset(0)

    def reset(self, burnin):
        self._counter = 0
        self._burnin = int(burnin)
        self._burnin >= 0

    def adapt_stepsize(self, accepted):
        """
        Adapts stepsize depending on whether the last attempt to generate a new state
        was successful (accepted=True) or not (accepted=False).
        """
        if self._counter < self._burnin:
            self.stepsize *= 1.02 if accepted else 0.98
            self._counter += 1


class MetropolisHastings(Sampler, AdaptiveStepsize):

    def __init__(self, distribution, initial_state, seed=None, stepsize=1e-1):
        """
        Parameters
        ----------
        distribution: instance of distributions.Distribution
            Target distribution implementing `log_prob`

        initial_state: numpy ndarray
            Initial state of the Markov chain.

        seed: int, optional
            Seed for the random number generator.

        stepsize: float > 0, optional
            Variance of isotropic Gaussian proposal.
        """
        super().__init__(distribution, initial_state, seed, stepsize=stepsize)
        self.n_accept = 0

    def propose(self):
        """
        Project into ambient space, perturb and project back to sphere.
        """
        x = np.copy(self.state)
        d = len(x)
        r = np.sqrt(2 * self.rng.gamma(d / 2))
        y = r * x + self.stepsize * self.rng.standard_normal(d)
        return sphere.radial_projection(y)

    def accept_reject(self, state):
        """
        Returns boolean indicating if new state can be accepted or should be
        rejected.
        """
        prob = self.target.log_prob(state) - self.target.log_prob(self.state)
        return np.log(self.rng.random()) < prob

    def __next__(self):
        state = self.propose()
        accepted = self.accept_reject(state)
        self.n_accept += int(accepted)
        self.adapt_stepsize(accepted)
        if accepted:
            self.state = state
        return np.copy(self.state)

    def sample(self, n_samples, burnin=0):
        """
        Generates a Markov chain consisting of the desired size. An optional
        burn-in can be specified, either as an integer (in which case it is
        taken as is) or a float smaller than one (in which case it is
        interpreted as percentage of the desired number of samples).
        """
        self.reset(determine_burnin(n_samples, burnin))
        return super().sample(n_samples, burnin)


class IndependenceSampler(MetropolisHastings):

    def propose(self):
        x = self.rng.standard_normal(len(self.state))
        return sphere.radial_projection(x)


def project(x, v):
    """
    Project x into complement of v. 
    """
    return x - v * (v @ x)


class SphericalHMC(MetropolisHastings):
    """
    Spherical Hamiltonian Monte Carlo.
    """

    def __init__(self, distribution, initial_state, seed=None, stepsize=1e-3, n_steps=10):
        """
        Parameters
        ----------
        distribution: instance of distributions.Distribution
            Target distribution implementing `log_prob` and `gradient`

        initial_state: numpy ndarray
            Initial state of the Markov chain.

        seed: int, optional
            Seed for the random number generator.

        stepsize: float > 0
            Stepsize of the leapfrog integrator.

        n_steps: integer > 0
            Number of leapfrog integration steps executed in proposal.
        """
        state = np.hstack([initial_state, np.zeros_like(initial_state)])
        super().__init__(distribution, state, seed, stepsize=stepsize)
        self.n_steps = int(n_steps)

    def sample_momenta(self):
        """
        Sample momenta in spherical HMC. 
        """
        x, v = np.reshape(self.state, (2, -1))
        v = project(self.rng.standard_normal(len(x)), x)
        self.state = np.hstack([x, v])

    def hamiltonian(self, state):
        """
        Evaluates Hamiltonian, i.e. kinetic energy minus log probability.
        """
        x, v = np.reshape(state, (2, -1))
        return 0.5 * v @ v - self.target.log_prob(x)

    def propose(self):
        """
        Run leapfrog integrator on sphere.
        """
        # log_prob = self.target.log_prob
        gradient = self.target.gradient
        eps = self.stepsize

        self.sample_momenta()

        x, v = np.reshape(np.copy(self.state), (2, -1))

        v += 0.5 * eps * project(gradient(x), x)

        for t in range(self.n_steps):

            norm = np.linalg.norm(v)

            y = np.copy(x)
            x = y * np.cos(eps * norm) + (v / norm) * np.sin(eps * norm)
            v = v * np.cos(eps * norm) - (y * norm) * np.sin(eps * norm)

            if t < self.n_steps - 1:
                v += eps * project(gradient(x), x)
            else:
                v += 0.5 * eps * project(gradient(x), x)

        return np.hstack([sphere.radial_projection(x), v])

    def accept_reject(self, state):
        prob = self.hamiltonian(self.state) - self.hamiltonian(state)
        return np.log(self.rng.random()) < prob

    def sample(self, n_samples, burnin=0, return_momenta=False):
        samples = super().sample(n_samples, burnin)
        positions, momenta = np.swapaxes(
            np.reshape(samples, (n_samples, 2, -1)), 1, 0
        )
        return (positions, momenta) if return_momenta else positions


class RejectionSphericalSliceSampler(Sampler):
    """
    Geodesic Spherical Rejection Slice Sampler
    """

    def __init__(self, distribution, initial_state, seed=None):
        """
        Parameters
        ----------
        distribution: instance of distributions.Distribution
            Target distribution implementing `log_prob` and `gradient`

        initial_state: numpy ndarray
            Initial state of the Markov chain.

        seed: int, optional
            Seed for the random number generator.
        """

        super().__init__(distribution, initial_state, seed)
        self.n_reject = 0

    def __next__(self):

        x = self.state
        p = self.target.log_prob

        # sample subsphere
        u = sphere.spherical_projection(self.rng.standard_normal(len(x)), x)

        threshold = p(x) + np.log(self.rng.random())

        interval = (0, 2 * np.pi)

        while True:
            theta = self.rng.uniform(*interval)
            y = np.cos(theta) * x + np.sin(theta) * u
            if p(y) > threshold:
                self.state = y
                return y
            self.n_reject += 1


class ShrinkageSphericalSliceSampler(RejectionSphericalSliceSampler):
    """
    Geodesic Spherical Shrinkage Slice Sampler
    """

    def __next__(self):

        x = self.state
        p = self.target.log_prob

        # sample subsphere
        u = sphere.spherical_projection(self.rng.standard_normal(len(x)), x)

        threshold = p(x) + np.log(self.rng.random())

        theta = self.rng.uniform(0, 2 * np.pi)
        interval = (theta - 2 * np.pi, theta)

        while True:
            theta = self.rng.uniform(*interval)
            y = np.cos(theta) * x + np.sin(theta) * u
            if p(y) > threshold:
                self.state = y
                return y
            interval = (theta, interval[1]) if theta < 0 \
                else (interval[0], theta)
            self.n_reject += 1




import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import Normalize

from geosss.distributions import Distribution

METHODS = ('sss-reject', 'sss-shrink', 'rwmh', 'hmc')
ALGOS = {'sss-reject': 'geoSSS (reject)', 'sss-shrink': 'geoSSS (shrink)',
         'rwmh': 'RWMH', 'hmc': 'HMC'}


def sphere_pdf(n_grid: int, pdf: Distribution):
    """ spherical grid with pdf"""

    # spherical grid
    u = np.linspace(0, np.pi, n_grid)
    v = np.linspace(0, 2 * np.pi, n_grid)

    u_grid, v_grid = np.meshgrid(u, v)
    vertices = np.stack([np.cos(v_grid) * np.sin(u_grid),
                        np.sin(v_grid) * np.sin(u_grid),
                        np.cos(u_grid)],
                        axis=2)

    # spherical to cartesian
    sph2cart = (
        np.outer(np.cos(v), np.sin(u)),
        np.outer(np.sin(v), np.sin(u)),
        np.outer(np.ones_like(u), np.cos(u))
    )

    # pdf values for the grid
    pdf_vals = np.array([np.exp(pdf.log_prob(val)) for val in vertices])

    return sph2cart, pdf_vals


def compare_samplers_3d(
        pdf: Distribution,
        samples: dict,
        n_grid: int = 100,
):

    for method in METHODS:
        msg = "visualization accepts only 3 dimension"
        assert samples[method].shape[1] == 3, msg

    sph2cart, pdf_vals = sphere_pdf(n_grid, pdf)
    pdfnorm = Normalize(vmin=pdf_vals.min(), vmax=pdf_vals.max())

    fig, axes = plt.subplots(2, 2, figsize=(8, 8), subplot_kw={
        "projection": "3d"}, sharex=True, sharey=True)

    for ax, method in zip(axes.flat, METHODS):
        ax.computed_zorder = False
        ax.plot_surface(*sph2cart,
                        facecolors=plt.cm.terrain_r(pdfnorm(pdf_vals)),
                        alpha=1, zorder=1)

        x = samples[method]
        ax.scatter(*x.T, c='tab:red', s=1, alpha=0.4, zorder=2)
        ax.set_title(ALGOS[method], pad=-30)
        ax.set_aspect('equal')
        ax.view_init(-140, 150)
        ax.axis('off')

    fig.tight_layout()
    plt.show()




from .demo_vis import compare_samplers_3d, sphere_pdf
from .distributions import (ACG, Bingham, BinghamFisher, Distribution,
                            MixtureModel, MultivariateNormal, VonMisesFisher,
                            random_bingham)
from .mcmc import (MetropolisHastings, RejectionSphericalSliceSampler,
                   ShrinkageSphericalSliceSampler, SphericalHMC)
from .rand import (sample_bingham, sample_bingham_2d, sample_bingham_3d,
                   sample_vMF)
from .sphere import (cartesian2polar, cartesian2spherical, distance, givens,
                     orthogonal_projection, polar2cartesian, radial_projection,
                     sample_sphere, sample_subsphere, spherical2cartesian,
                     spherical_projection)
from .utils import (IAT, SamplerLauncher, acf, acf_fft, colors, counter, n_eff,
                    take_time)



# diagnostics and plots for mixture of vMF as target (although mostly for general targets,
# but need to check that)

import os

import arviz as az
import matplotlib.pyplot as plt
import numpy as np
from csb.io import dump, load

from geosss.sphere import distance
from geosss.utils import acf

METHODS = ['sss-reject', 'sss-shrink', 'rwmh', 'hmc']
ALGOS = {'sss-reject': 'geoSSS (reject)',
         'sss-shrink': 'geoSSS (shrink)',
         'rwmh': 'RWMH',
         'hmc': 'HMC'}

plt.rc("font", size=16)


def hist_plot(samples, ndim, path, filename, fs=16, save_res=False):
    """ 
    histogram of samples. 
    """
    bins = 100
    plt.rc("font", size=fs)

    # shows a standard histogram per dimension
    if ndim == 3:
        figsize = (10, 10)
    else:
        figsize = (10, 15)
    fig, rows = plt.subplots(ndim, len(METHODS), figsize=figsize,
                             sharex=True)

    for i, axes in enumerate(rows):

        # reference samples
        wood_vals = samples['wood'][:, i]
        ref = list(np.histogram(wood_vals, bins=bins, density=True))
        ref[1] = 0.5 * (ref[1][1:] + ref[1][:-1])

        # show histogram
        for ax, method in zip(axes, METHODS):
            marginals = samples[method][:, i]

            bins = ax.hist(marginals, bins=bins, density=True, alpha=0.3,
                           color='k', histtype='stepfilled')[1]
            ax.plot(*ref[::-1], color='r', lw=1, ls='--')
            ax.set_xlabel(fr'$e_{i}^Tx_n$', fontsize=fs)

    for ax, method in zip(rows[0], METHODS):
        ax.set_title(ALGOS[method], fontsize=fs)
    fig.tight_layout()

    if save_res:
        print(f"Saving sampler marginals plot..")
        fig.savefig(f"{path}/{filename}_hist.pdf", transparent=True)

    plt.close(fig)


def trace_plots(samples, ndim, path, filename, fs=16, save_res=False):
    """ 
    trace plots per dimension
    """

    os.makedirs(f"{path}/trace_plots", exist_ok=True)

    for d in range(ndim):
        fig, axes = plt.subplots(1, len(METHODS), figsize=(
            12, 5), sharex=True, sharey=True)

        for ax, method in zip(axes, METHODS):
            ax.set_title(ALGOS[method], fontsize=fs)
            ax.plot(samples[method][:, d], alpha=0.5, color='k', lw=1)
            ax.set_xlabel(r'MCMC step $n$', fontsize=fs)
        axes[0].set_ylabel(r'$u_{d}^Tx_n$', fontsize=fs)
        fig.suptitle(fr"Trace plot $d_{{{d+1}}}$")
        if save_res:
            print(f"saving trace plot for dimension {d+1}..")
            fig.savefig(
                f"{path}/trace_plots/{filename}_trace_d{d+1}.pdf", transparent=True)

        plt.close(fig)


def acf_plots(samples, ndim, path, filename, lag=1000, fs=16, save_res=False):
    """
    Plots acf for every dimension in a multidimensional target
    """

    os.makedirs(f"{path}/acf_plots", exist_ok=True)

    for d in range(ndim):
        fig, ax = plt.subplots(1, 1, figsize=(6, 5))

        for method in METHODS:
            ac = acf(samples[method][:, d], lag)
            ax.plot(ac, alpha=0.7, lw=3, label=ALGOS[method])
            ax.legend(fontsize=fs)
            ax.axhline(0., ls='--', color='k', alpha=0.7)
            ax.set_xlabel(r'Lag', fontsize=fs)
            ax.set_ylabel('ACF', fontsize=fs)
            ax.set_title(fr"$d_{{{d+1}}}$")
        if save_res:
            print(f"saving acf plots for dim {d+1}..")
            fig.savefig(
                f"{path}/acf_plots/{filename}_acf_d{d+1}.pdf", transparent=True)

        plt.close(fig)


def acf_entropy_plot(samples, pdf, path, filename, lag=1000, fs=16, save_res=False):
    """ 
    ACF entropy plot 
    """

    # population of modes
    modes = np.array([p.mu for p in pdf.pdfs])

    # kl-divergence and entropy
    KL = []
    H = []
    for method in METHODS:
        x = samples[method]
        m = np.argmax(x @ modes.T, axis=1)
        i, c = np.unique(m, return_counts=True)
        p = np.full(len(modes), 1e-100)
        p[i] = c
        p[i] /= p.sum()
        KL.append(p @ np.log(p / pdf.weights))
        H.append(-(p @ np.log(p)))

    fig, axes = plt.subplots(1, 2, figsize=(10, 4))
    ax = axes[0]
    for method in METHODS:
        ac = acf(samples[method][:, 0], lag)
        ax.plot(ac, alpha=0.7, lw=3, label=ALGOS[method])
    ax.legend(fontsize=fs)
    ax.axhline(0., ls='--', color='k', alpha=0.7)
    ax.set_xlabel(r'Lag', fontsize=fs)
    ax.set_ylabel('ACF', fontsize=fs)

    # hopping frequency as bar plot
    ax = axes[1]
    ax.set_ylabel("Entropy")
    ax.bar(list(map(ALGOS.get, METHODS)), H, color='k', alpha=0.3)
    ax.axhline(np.log(len(pdf.pdfs)), ls='--', color='r', label='max entropy')
    plt.xticks(rotation=30)
    fig.tight_layout()

    if save_res:
        print(f"Saving ACF-Entropy plot..")
        fig.savefig(f"{path}/{filename}_acf_entropy.pdf", transparent=True)

    plt.close(fig)


def entropy_kld(samples, pdf, path, filename, save_res=False):
    """ 
    Plots entropy and KL divergence plot between 
    target and sampled distribution
    """

    # population of modes
    modes = np.array([p.mu for p in pdf.pdfs])
    KL = []
    H = []
    for method in METHODS:
        x = samples[method]
        m = np.argmax(x @ modes.T, axis=1)
        i, c = np.unique(m, return_counts=True)
        p = np.full(len(modes), 1e-100)
        p[i] = c
        p[i] /= p.sum()
        KL.append(p @ np.log(p / pdf.weights))
        H.append(-(p @ np.log(p)))

    fig, axes = plt.subplots(1, 1, figsize=(6, 4))
    ax = axes
    ax.set_ylabel("KL divergence")
    ax.bar(list(map(ALGOS.get, METHODS)), KL, color='k', alpha=0.3)
    plt.xticks(rotation=30)
    fig.tight_layout()
    if save_res:
        print(f"Saving KL divergence plot..")
        fig.savefig(f"{path}/{filename}_kl.pdf",
                    bbox_inches='tight', transparent=True)

    fig, axes = plt.subplots(1, 1, figsize=(6, 4))
    ax = axes
    ax.set_ylabel("Entropy")
    ax.bar(list(map(ALGOS.get, METHODS)), H, color='k', alpha=0.3)
    plt.xticks(rotation=30)
    fig.tight_layout()
    if save_res:
        print(f"Saving entropy plot..")
        fig.savefig(f"{path}/{filename}_entropy.pdf",
                    bbox_inches='tight', transparent=True)

    plt.close(fig)


def dist_plot(samples, pdf, kappa, path, filename, fs=16, save_res=False):
    """ 
    plotting Geodesic distance
    """

    modes = np.array([p.mu for p in pdf.pdfs])
    d_modes = np.concatenate(
        [distance(modes/kappa, mode/kappa) for mode in modes])
    d_modes = np.unique(np.round(d_modes, 5))
    fig, axes = plt.subplots(1, len(METHODS), figsize=(len(METHODS) * 3, 3),
                             sharex=True, sharey=True)
    bins = 100
    for ax, method in zip(axes, METHODS):
        ax.set_title(ALGOS[method], fontsize=fs)

        # distance between successive samples
        x = samples[method]
        d = distance(x[:-1], x[1:])
        print('average great circle distance of successive samples: '
              f'{np.mean(d):.2f} ({method})')

        # plot distance as histogram
        bins = ax.hist(d, bins=bins, density=True, alpha=0.3,
                       color='k', histtype='stepfilled')[1]
        for d_mode in d_modes[:0]:
            ax.axvline(d_mode, ls='--', color="r")
        ax.set_xlabel(r'$\delta(x_{n+1}, x_n)$', fontsize=fs)
        ax.set_xticks(np.linspace(0., np.pi, 3))
        ax.set_xticklabels(["0", r"$\pi/2$", r"$\pi$"])
        ax.semilogy()
    fig.tight_layout()

    if save_res:
        print(f"Saving geodesic distance plot..")
        fig.savefig(f"{path}/{filename}_dist.pdf", transparent=True)

    plt.close(fig)


def calc_ess(runs_samples, methods, path, return_ess=True):
    """ 
    calculates ess using the arviz library with the default 'bulk' method
    and saves the result. This implementation implements for multidimensional 
    target and estimates ess values per dimension using `n` chains 
    """
    ess_file = f"{path}_ess.pkl.gz"

    # load or calculate ess (and then save)
    try:
        ess = load(ess_file, gzip=True)
        print(f"Loading ESS file {ess_file}")

    except FileNotFoundError:

        # calculate ess when `n_runs=10`
        if isinstance(runs_samples, list):
            if len(runs_samples) == 10:

                print(f"Calculating ESS from samples..")
                ess = {method: None for method in methods}

                for method in methods:

                    # samples from all runs with shape (chains, draws, dimensions)
                    samples = np.array([draws[method]
                                       for draws in runs_samples])

                    # estimates ESS using the arviz library per dimension
                    samples_az = az.convert_to_dataset(samples)
                    ess_dims = az.ess(samples_az, relative=True)
                    ess[method] = ess_dims.x.values

                print(f"Saving ESS file {ess_file}")
                dump(ess, ess_file, gzip=True)

            else:
                print(f"ESS values not computed, requires `n_runs=10`")
                return None

    for method in methods:
        for i, vals in enumerate(ess[method]):
            print(f"{method} ESS dim {i+1}: {vals:.8%}")

    if return_ess:
        return ess
